from collections import defaultdict
import os
from pathlib import Path

import git
from pydantic import ValidationError
import toml

from .configuration_sources import StandardConfiguration
import typer
import tomli_w

environment_app = typer.Typer(help="Manage and investigate environments")


@environment_app.command()
def configuration():
    """
    Display the credential loading process and final resolved credentials.

    This command shows the sequence of credential loaders that were attempted,
    the configuration layers they produced, and the final merged credentials
    that will be used by the application. The credentials are displayed in
    TOML format with environments shown separately from other credential fields.
    The output includes:

        * Each credential loader attempted and its resulting layers

        * Final merged credentials (excluding environments) in green

        * Environment configurations displayed separately
    """

    configuration = StandardConfiguration()
    typer.secho("Credential Loaders Found:\n")
    for loader, layer in zip(configuration.loaders, configuration.layers):
        if layer:
            typer.secho(str(loader))
            typer.secho("-" * len(str(loader)))

            typer.secho(toml.dumps(layer))
            typer.secho()

    typer.secho("Final configuration")
    typer.secho("=================\n")

    typer.secho(
        tomli_w.dumps(
            {k: v for k, v in configuration.items().items() if k != "environments"}
        ),
        fg="green",
    )

    try:
        configuration.validate()
    except ValidationError as e:
        typer.secho(e, fg="red")
        exit()


@environment_app.command()
def list():
    """
    List all environments.
    """

    configuration = StandardConfiguration()
    current_environment = configuration.environment

    for k in configuration.known_environments():
        if k != current_environment:
            print(f"  {k}")
        else:
            typer.secho(f"* {k}", fg="green")


@environment_app.command()
def select(environment: str):
    """
    Select an environment
    """
    configuration = StandardConfiguration()

    if current_environment := os.environ.get("RELY_ENVIRONMENT"):
        typer.secho(
            f"Environment variable RELY_ENVIRONMENT={current_environment} is set. Changes will have no effect.",
            fg="red",
        )
        exit()
    else:
        typer.secho(f"Current environment: {configuration.environment}")

    known_environments = configuration.known_environments()
    if environment not in known_environments:
        typer.secho(
            f"The selected environment {environment} does not have any known configurations, the known environments are:",
            fg="red",
        )
        for known_env in known_environments:
            typer.secho("  * " + known_env, fg="red")
        if not typer.confirm("Are you sure you wish to continue"):
            exit()

    # Find closest .rely.toml
    environment_file_path = None
    for root in configuration.directory_ladder():
        path = root / ".rely-environment"
        if path.exists():
            typer.secho(f"Found {path}")
            environment_file_path = path
            break

    # Otherwise offer to create one in the git root
    if not environment_file_path:
        repo = git.Repo(os.getcwd(), search_parent_directories=True)
        environment_file_path = Path(repo.working_dir) / ".rely-environment"
        if not typer.confirm(
            f"No .rely-environment found, would you like to create one at the git root ({environment_file_path})"
        ):
            typer.secho("Environment could not be set", fg="red")
            exit()

    with open(environment_file_path, "w") as f:
        f.write(environment)
        typer.secho(f"Environment set to {environment}", fg="green")



from .configuration_sources import ConfigurationLoader


class VariableComputer:
    def __init__(self, configuration: ConfigurationLoader):
        self.configuration = configuration

    def variables(self):
        return self.configuration.items()



import os
from typing import Any
from pydantic import ValidationError
from relycomply_client.types import ConfigurationModel
import toml
from pathlib import Path
import logging
import git
import os
import boto3
import base64
from botocore.exceptions import ClientError
import typer


from .exceptions import RelyComplyClientException

log = logging.getLogger(__name__)
keys = ["token", "url", "impersonate", "variables"]


def dict_merge(dct, merge_dct):
    """Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.
    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None
    """
    for k in merge_dct.keys():
        if k in dct and isinstance(dct[k], dict) and isinstance(merge_dct[k], dict):  # noqa
            dict_merge(dct[k], merge_dct[k])
        else:
            dct[k] = merge_dct[k]


def merge_credentials(layers):
    credentials = {}
    for layer_credentials in layers:
        dict_merge(credentials, layer_credentials)
    return credentials


class ConfigurationLoader:
    def __str__(self):
        return type(self).__name__ + "()"

    def items(self) -> dict[str, Any]: ...

    def variables(self) -> dict[str, Any]:
        return self.get("variables", {})  # type: ignore

    def get(self, key, default=None):
        return self.items().get(key, default)

    def require(self, key):
        value = self.get(key)
        if not value:
            raise RelyComplyClientException(
                f"Configuration '{key}' is required but not present"
            )
        return value


class Default(ConfigurationLoader):
    def items(self):
        return {"url": "https://app.relycomply.com"}


class Environment(ConfigurationLoader):
    def items(self):
        return {
            key: os.environ[f"RELYCOMPLY_{key.upper()}"]
            for key in keys
            if f"RELYCOMPLY_{key.upper()}" in os.environ
        }


class ConfigFolder(ConfigurationLoader):
    def __init__(self, folder: Path, filename=".rely.toml"):
        self.filename = filename
        self.folder = folder

    def items(self):
        config_path = self.folder / self.filename
        if not config_path.exists():
            return {}
        else:
            with open(config_path) as f:
                config_values = toml.load(f)
                return config_values

    def __str__(self):
        return type(self).__name__ + f"({self.folder}/{self.filename})"


class SecretsManager(ConfigurationLoader):
    def load_secret(self, key):
        session = boto3.session.Session()  # type: ignore
        client = session.client(
            service_name="secretsmanager",
        )
        secret_name = key

        try:
            get_secret_value_response = client.get_secret_value(SecretId=secret_name)
        except ClientError as e:
            raise RelyComplyClientException(
                f"Could not retrieve secret {key}: str(e)"
            ) from e
        else:
            # Decrypts secret using the associated KMS CMK.
            # Depending on whether the secret is a string or binary, one of these fields will be populated.
            if "SecretString" in get_secret_value_response:
                return get_secret_value_response["SecretString"]
            else:
                return base64.b64decode(get_secret_value_response["SecretBinary"])

    def items(self):
        secret_keys = {
            key: os.environ[f"RELYCOMPLY_{key.upper()}_AWS_SECRET"]
            for key in keys
            if f"RELYCOMPLY_{key.upper()}_AWS_SECRET" in os.environ
        }
        return {key: self.load_secret(value) for key, value in secret_keys.items()}


class Arguments(ConfigurationLoader):
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def items(self):
        return {key: self.kwargs[key] for key in keys if self.kwargs.get(key)}


class StandardConfiguration(ConfigurationLoader):
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self._values = None

        environment = self.environment
        self.loaders = [
            Default(),
            *self.folder_loaders(environment),
            SecretsManager(),
            Environment(),
            Arguments(**self.kwargs),
        ]
        log.info("Loading Credentials")
        self.layers = [loader.items() for loader in self.loaders]
        self.results = merge_credentials(self.layers)

        for loader, layer in zip(self.loaders, self.layers):
            log.info(f"{loader}: {layer}")

        log.info(f"Merged Credentials: {self.results}")

        # TODO: Branch protection

        log.info(f"Final Credentials: {self.results}")

    def directory_ladder(self):
        root = Path(".").resolve()
        yield root
        yield from root.parents

    def folder_loaders(self, environment: str | None):
        folder_credentials = []

        file_patterns = (
            [
                ".rely.toml",
                f".rely-configuration-{environment}.toml",
                ".rely-configuration.toml",
                f"rely-configuration-{environment}.toml",
                "rely-configuration.toml",
            ]
            if environment
            else [
                ".rely.toml",
                ".rely-configuration.toml",
                "rely-configuration.toml",
            ]
        )
        for root in self.directory_ladder():
            for filename in file_patterns:
                if (root / filename).exists():
                    folder_credentials.append(ConfigFolder(root, filename))

        return reversed(folder_credentials)

    @property
    def environment(self):
        if not hasattr(self, "_environment"):
            self._environment = self.find_environment()
        return self._environment

    def find_environment(self):
        if environment := os.environ.get("RELY_ENVIRONMENT"):
            return environment

        for root in self.directory_ladder():
            if (root / ".rely-environment").exists():
                with open(root / ".rely-environment") as f:
                    environment = f.read()
                    return environment

    def known_environments(self):
        environments = set()
        for root in self.directory_ladder():
            for x in root.glob(
                "*rely-configuration-*.toml",
            ):
                environment = x.stem.split("-")[-1]
                environments.add(environment)
        return environments

    def items(self):
        return self.results

    @property
    def values(self):
        if not self._values:
            self._values = ConfigurationModel.model_validate(self.results)

        return self._values

    def validate(self):
        try:
            ConfigurationModel.model_validate(self.results)
        except ValidationError as e:
            typer.secho(
                f"The provided configuration contains {e.error_count()} validation errros",
                fg="red",
            )

            for error in e.errors():
                path = ".".join([str(i) for i in error["loc"]])
                typer.secho(f"  * {path}: {error['msg']}", fg="red")
            print()
            typer.secho(
                "Please run `rely environment configuration` for more information and diagnostics.",
                fg="red",
            )
            exit()



import sys
import typer
from relycomply_client.configuration_sources import StandardConfiguration
from relycomply_client.gql_client import RelyComplyGQLClient
from graphql import print_schema


def introspect():
    """
    Introspect the GraphQL schema and print it to stdout
    """
    configuration = StandardConfiguration()
    configuration.validate()
    gql_client = RelyComplyGQLClient(configuration=configuration)
    if gql_client.schema:
        print(print_schema(gql_client.schema))
    else:
        print("Could not retrieve schema.", file=sys.stderr)
        raise typer.Exit(code=1)



from relycomply_client.configuration_sources import StandardConfiguration
from relycomply_client.gql_client import RelyComplyGQLClient
from .sync import sync
from .render import render
from .watch import watch_sync
from .cli import RelyComplyCLI
from .environment import environment_app
import typer
from .turbo import turbo, turbo_help_text
from .introspect import introspect

help_text = """
The RelyComply client
"""

app = typer.Typer(rich_markup_mode="markdown", help=help_text)

app.command()(sync)
app.command("watch")(watch_sync)
app.command()(render)
app.command(hidden=True)(introspect)


def version_callback(value: bool):
    if value:
        import importlib.metadata

        __version__ = importlib.metadata.version("relycomply_client")
        typer.echo(__version__)
        raise typer.Exit()


@app.callback()
def common(
    ctx: typer.Context,
    version: bool = typer.Option(
        None,
        "--version",
        callback=version_callback,
        help="Show the current version of the relycomply client",
    ),
):
    pass


@app.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def cli(ctx: typer.Context):
    """
    A convinient CLI for interacting with the RelyComply GraphQL API
    """
    configuration = StandardConfiguration()
    configuration.validate()
    gql_client = RelyComplyGQLClient(configuration=configuration)
    cli = RelyComplyCLI(gql_client=gql_client)
    cli.run_command(ctx.args)


app.add_typer(environment_app, name="environment")

app.command(help=turbo_help_text())(turbo)


def main():
    app()
